/* System headers */
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <assert.h>


/* The name of the implementation */
#define NAME  "libdynamic"


/* The implementation (in terms of inline inclusion of source code) */
#include "dynamic/map.c"
#include "dynamic/hash.c"


/* Inline utilities to deal with objects for operations (put/get/delete/iterate) in the hash tables */
#include "objects.c"


/*
 * The set callback is called with a pointer to a source element from where
 * the element data is read, and a destination element where the data is written.
 */
static void my_set (map * map, void * _dst, void * _src)
{
  robj_t ** dst = _dst;
  robj_t ** src = _src;

  if (src)
    * dst = * src;
}


/*
 * The hash callback is called with a pointer a map element and should return a hash value of the key of the element.
 */
static size_t my_hash (map * ht, void * _obj)
{
  robj_t ** obj = _obj;
  return hash_string ((* obj) -> skey);
}


/*
 * The equal callback is called with a pointer to two elements, element1 and element2,
 * and should return 1 if the elements are equal.
 */
static int my_equal (map * ht, void * _obj1, void * _obj2)
{
  robj_t * obj1 = _obj1 ? * (robj_t **) _obj1 : NULL;
  robj_t * obj2 = _obj2 ? * (robj_t **) _obj2 : NULL;

  return obj1 == obj2 || (obj1 && obj2 && obj1 -> skey && obj2 -> skey && ! strcmp (obj1 -> skey, obj2 -> skey));
}


/* =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=- */


/* Allocate and populate a hash table with n objects */
static map * grow (unsigned n, robj_t * objs [])
{
  map * ht = calloc (1, sizeof (* ht));
  unsigned i;

  assert (ht);
  assert (map_size (ht) == 0);

  /* Hash table initialization */
  map_construct (ht, sizeof (robj_t *), NULL, my_set);        /* use null pointer as empty object */

  /* Insert objects into the hash table */
  for (i = 0; i < n; i ++)
    {
      map_insert (ht, & objs [i], my_hash, my_equal, my_set, NULL);
      assert (map_size (ht) == i + 1);
    }
  assert (map_size (ht) == n);

  return ht;
}


/* === Unit tests === */


/* Basic allocate and free */
static void alloc_free (void)
{
  map * ht = calloc (1, sizeof (* ht));
  free (ht);
}


/* Basic allocate, init and free */
static void alloc_init_free (unsigned n)
{
  map * ht = calloc (1, sizeof (* ht));

  /* Hash table initialization */
  map_construct (ht, sizeof (robj_t *), NULL, my_set);        /* use null pointer as empty object */

  /* Hash table and memory cleanup */
  map_destruct (ht, NULL, NULL);
  free (ht);
}


/* Basic insert */
static void alloc_insert_free (unsigned n)
{
  /* Allocate memory to keep all the objects (unique serial initialization) */
  robj_t ** objs = mkobjargv (n);
  map * ht = grow (n, objs);

  /* Hash table and memory cleanup */
  map_destruct (ht, NULL, NULL);
  free (ht);
  rmobjargv (objs, n);
}


static unsigned alloc_insert_count_free (unsigned n)
{
  /* Allocate memory to keep all the objects (unique serial initialization) */
  robj_t ** objs = mkobjargv (n);
  map * ht = grow (n, objs);
  unsigned count = map_size (ht);

  /* Hash table and memory cleanup */
  map_destruct (ht, NULL, NULL);
  free (ht);
  rmobjargv (objs, n);

  return count;
}


static void alloc_insert_delete_free (unsigned n)
{
  /* Allocate memory to keep all the objects (unique serial initialization) */
  robj_t ** objs = mkobjargv (n);
  map * ht = grow (n, objs);
  unsigned i;

  for (i = 0; i < n; i ++)
    map_erase (ht, & objs [i], my_hash, my_equal, my_set, NULL);
  assert (map_size (ht) == 0);

  /* Hash table and memory cleanup */
  map_destruct (ht, NULL, NULL);
  free (ht);
  rmobjargv (objs, n);
}


/* Basic insert and attempt to remove not existing */
static void alloc_insert_missed_free (unsigned n)
{
  /* Allocate memory to keep all the objects (unique serial initialization) */
  robj_t ** objs = mkobjargv (n);
  map * ht = grow (n, objs);
  unsigned i;

  for (i = 0; i < n; i ++)
    {
      robj_t * missing = mkobj (n + 1, n + 1);
      map_erase (ht, & missing, my_hash, my_equal, my_set, NULL);
      rmobj (missing);
      assert (map_size (ht) == n);
    }
  assert (map_size (ht) == n);

  /* Hash table and memory cleanup */
  map_destruct (ht, NULL, NULL);
  free (ht);
  rmobjargv (objs, n);
}


static void alloc_insert_found_free (unsigned n)
{
  /* Allocate memory to keep all the objects (unique serial initialization) */
  robj_t ** objs = mkobjargv (n);
  map * ht = grow (n, objs);
  unsigned i;

  for (i = 0; i < n; i ++)
    assert (map_at (ht, & objs [i], my_hash, my_equal));

  /* Hash table and memory cleanup */
  map_destruct (ht, NULL, NULL);
  free (ht);
  rmobjargv (objs, n);
}


static void alloc_insert_notfound_free (unsigned n)
{
  /* Allocate memory to keep all the objects (unique serial initialization) */
  robj_t ** objs = mkobjargv (n);
  map * ht = grow (n, objs);
  unsigned i;

  for (i = 0; i < n; i ++)
    {
      robj_t * missing = mkobj (n + 1, n + 1);
      assert (map_at (ht, & missing, my_hash, my_equal));
      rmobj (missing);
    }

  /* Hash table and memory cleanup */
  map_destruct (ht, NULL, NULL);
  free (ht);
  rmobjargv (objs, n);
}


#if defined(ROCCO)
/* Basic insert and iterate */
static unsigned alloc_insert_iterate_free (unsigned n)
{
  robj_t ** objs = mkobjargv (n);
  map * ht = grow (n, objs);
  unsigned count = 0;
  unsigned i;

  for (i = 0; i < ht -> elements_capacity; i ++)
    count ++;
  assert (count == n);

  /* Hash table and memory cleanup */
  map_destruct (ht, NULL, NULL);
  free (ht);
  rmobjargv (objs, n);

  return count;
}
#endif /* ROCCO */

int main (int argc, char * argv [])
{
  unsigned n = argc == 1 ? DEFAULT_OBJS : atoi (argv [1]);

  if (! n)
    n = DEFAULT_OBJS;

  printf ("%s: basic testsuite (#%u)\n", NAME, n);

  printf ("Obj .......................... ");
  rmobj (mkobj (n, n));
  printf ("Ok\n");

  printf ("Objs ......................... ");
  rmobjs (mkobjs (n), n);
  printf ("Ok\n");

  printf ("Alloc/Free ................... ");
  alloc_free ();
  printf ("Ok\n");

  printf ("Alloc/Init/Free .............. ");
  alloc_init_free (n);
  printf ("Ok\n");

  printf ("Alloc/Insert/Free ............ ");
  alloc_insert_free (n);
  printf ("Ok\n");

  printf ("Alloc/Insert/Count/Free ...... ");
  printf ("Ok %u\n", alloc_insert_count_free (n));

  printf ("Alloc/Insert/Delete/Free ..... ");
  alloc_insert_delete_free (n);
  printf ("Ok\n");

  printf ("Alloc/Insert/Missed/Free ..... ");
  alloc_insert_missed_free (n);
  printf ("Ok\n");

  printf ("Alloc/Insert/Found/Free ...... ");
  alloc_insert_found_free (n);
  printf ("Ok\n");

  printf ("Alloc/Insert/NotFound/Free ... ");
  alloc_insert_notfound_free (n);
  printf ("Ok\n");

  printf ("Alloc/Insert/Iterate/Free .... ");
#if defined(ROCCO)
  printf ("Ok %u\n", alloc_insert_iterate_free (n));
#else
  printf ("Not yet available\n");
#endif /* ROCCO */

  return 0;
}
